/**************************************************************************************

 disass.c
 
 Description:
 
 	Disassembler module for Asm45
                           
 History:

	Version 1.1:		- Added base page constants & variables plus comments

	Version 1.0 alpha:	First release

 License information:

	Copyright (c) 2009 Ansgar Kueckes

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 
***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>
#include <math.h>
#include "../include/asm45.h"

/* == disassembler & analyzer routines =============================================== */

int search_label (Label *list, int size, unsigned int address)
{
	int i;
	
	for (i=0; i < size; i++) {
		if (list[i].address == address) break;
	}
	if (i == size) return 0;
	return i+1; /* found */
}

void sort_label_list(Label *list, int size)
{
	int i,j,k;
	unsigned int tmp_address = 0;
	unsigned int tmp_value = 0;
	int tmp_reftype[10];
	int tmp_refcount;
	Node *tmp_ref[10];
	
	for (i=0; i < size; i++)
	    for (j=i; j < size; j++)
	        if (list[i].address > list[j].address) {
	            tmp_address = list[i].address;
	            tmp_value = list[i].value;
	            tmp_refcount = list[i].refcount;
	            for (k=0;k<10;k++) {
	            	tmp_reftype[k] = list[i].reftype[k];
	                tmp_ref[k] = list[i].ref[k];
	            }
	            
	            list[i].address = list[j].address;
	            list[i].value = list[j].value;
	            list[i].refcount = list[j].refcount;
	            for (k=0;k<10;k++) {
	            	list[i].reftype[k] = list[j].reftype[k];
	            	list[i].ref[k] = list[j].ref[k];
	            }
	            
	            list[j].address = tmp_address;
	            list[j].value = tmp_value;
	            list[j].refcount = tmp_refcount;
	            for (k=0;k<10;k++) {
	            	list[j].reftype[k] = tmp_reftype[k];
	            	list[j].ref[k] = tmp_ref[k];
	            }
	        }
}

void update_label(Label *list, int *size, int type, unsigned int address, unsigned int target)
{
    int i,label;
    Node *node,*tmp;

    label = search_label(list, *size, target);
    
    if (label == 0) { // new label
        *size=*size+1;
        label = *size;
        if (label > (MAX_LABEL-1)) {
            printf("Sorry, max. label limit (%d labels) exceeded.\n", MAX_LABEL);
            exit (-1);
        }
	  	list[label-1].refcount = 0;
	  	for (i=0; i<10; i++) {
	  		list[label-1].reftype[i] = 0;
	  		list[label-1].ref[i] = 0;
	        }
	    list[label-1].address = target;
	}
    node = list[label-1].ref[type];
    if (node) {
        while (node->next) node = node->next;
        node->next = malloc(sizeof(Node));
        tmp = node;
        node = node->next;
    }
    else {
		node = malloc(sizeof(Node));
		list[label-1].ref[type] = node;
		tmp = 0;
    }
    node->address = address;
    node->next = 0;
    list[label-1].reftype[type]++;
    list[label-1].refcount++;
}

void disassemble(int line, unsigned int address, unsigned int cmd, FILE *ofp)
{
	unsigned int operand;
	Op_code *current = op_table;
	char asc[3];
	int i,j;
	int target = 0;
	int label = 0;
	int timing;
	int tmp;
	char out[1024];   // output string for disassembler

	strcpy(out,"");
	
	if (numbers) printf("%3d ", line);
	
	if (ascii) {
	    i = cmd / 256;
	    if (i>32 && i<127) asc[0] = i; else asc[0] = ' ';
	    i = cmd - (i * 256);
	    if (i>32 && i<127) asc[1] = i; else asc[1] = ' ';
	    asc[2] = 0;
	    printf_string(out,"'%s' ", asc);
	}
	
	if (isource) printf_string(out, "ISOURCE ");
	else {
		if (octal)
	        printf_string(out,"%06o %06o ", address, cmd);
		else
		    printf_string(out,"%04x %04x ", address, cmd);
	}
	label = search_label(internals, internals_size, address);
	if (label)
	    printf_string(out,"Lbl_%04i: ", label);
	else
	    printf_string(out,"          ");
	
	for (i=0; i < op_table_size; i++, current++){
	    if ((cmd & current->op_mask) == current->pattern) {
	    	operand = cmd & ~(current->op_mask);
	    	
	    	printf_string(out,"%s ", current->opcode);
	    	timing = current->timing;
	    	label=0;
	    	
	    	switch (current->type) {
	    	    case 0:  /* no operand */
	    	             break;
	    	    
	    	    case 1:  /* location */
	    	    case 2:  /* indirect location */
	    	    	     if (operand & 0x400) { /* current page addressing */
	    	    	        operand = operand & 0x3ff;
	    	    	     	if (operand < 512) target = address + operand;
	    	    	     	else target = address + operand - 1024 ;
	    	    	     	if (target>65535) target-=65536;
	    	    	     	else if (target<0) target+=65536;
	    	    	        target &= 0xFFFF;
	    	    	     	if (target == address) {
	    	    	     	    printf_string(out,"*");
	    	    	     	}
	    	    	     	else {
	    	    	     	    label = search_label(internals, internals_size, target);
				    			if (label)
	                                 printf_string(out,"Lbl_%04i", label);
	                            else {
	    	     	             	if (octal) printf_string(out,"%06oB", target);
    	    	     	            else printf_string(out,"$%04x", target);
	    	    	     	    }
	    	    	     	}

	    	    	     	if (current->type == 2)
	    	    	     		printf_string(out,",I"); /* indirect marker */

							if (label) { /* reference for memory access or indirect JMP/JSR */
				    				tmp = (cmd >> 11) & 0x1F;
                                    if (((tmp&0xF)<0x6) ||	// LDx, CPx, ADx,
                                        ((tmp&0x1E)==0x16) ||	// indirect STx
                                        ((tmp&0xF)==0xA) ||	// AND
                                        ((tmp&0xF)==0xC) ||	// IOR
                                        (tmp==0x18) ||	// indirect JSM
                                        (tmp==0x1D)) {	// indirect JMP
                                        if (!isource) {
	    	    	     	                if (octal) {
	    	    	     	                	printf_string(out," [=%06oB]", internals[label-1].value);
	    	    	     	                }
	    	    	     	                else {
	    	    	     	                	printf_string(out," [=$%04x]", internals[label-1].value);
	    	    	     	                }
    	    	     	            	}
                                    }
							}
	    	    	     } else { /* base page addressing */
	    	    	     	operand = operand & 0x3ff;
	    	    	     	if (operand < 32) {
	    	    	     		printf_string(out,"%s", predefined[operand].name); /* register */
	    	    	     	}
	    	    	     	else {
	    	    	     	    if (operand < 512) target = operand;
	    	    	     	    else target = 65536 - 1024 + operand;
	    	    	            target &= 0xFFFF;
	    	    	            label = 0;
	    	    	     	    if (base_page_mode) label = search_label(internals, internals_size, target);
					    		if (label)
		                                 printf_string(out,"Lbl_%04i", label);
	                            else {
	                            	 if (map_basepage && (target >= 0x20) && (target < 0x0200)) {
	                            	 	if (*rom_labels[target]) {
	                            	 		printf_string(out, "%s", rom_labels[target]);
	                            	 		//rom_labels_access[target]++;
	                            	 	}
	                            	 	else {
	    	    	     	             	if (octal) printf_string(out,"%06oB", target);
		    	    	     	            else printf_string(out,"$%04x", target);
	                            	 	}
	                            	 }
	                            	 else if (map_basepage && (target >= 0xfe00)) {
	                            	 	if (*ram_labels[target-0xfe00]) {
	                            	 		printf_string(out, "%s", ram_labels[target-0xfe00]);
	                            	 		//ram_labels_access[target-0xfe00]++;
	                            	 	}
	                            	 	else {
	    	    	     	             	if (octal)printf_string(out,"%06oB", target);
		    	    	     	            else printf_string(out,"$%04x", target);
	                            	 	}
	                            	 }
    	    	     	             else {
    	    	     	             	if (octal) printf_string(out,"%06oB", target);
	    	    	     	            else printf_string(out,"$%04x", target);
    	    	     	            }
	    	    	     	    }
							}
							
	    	    	     	if (current->type == 2)
	    	    	     		printf_string(out,",I");

							if (label || map_basepage) { /* reference for memory access or indirect JMP/JSR */
				    			tmp = (cmd >> 11) & 0x1F;
                                if (((tmp&0xF)<0x6) ||	// LDx, CPx, ADx
                                    ((tmp&0x1E)==0x16) ||	// indirect STx
                                    ((tmp&0xF)==0xA) ||	// AND
                                    ((tmp&0xF)==0xC) ||	// IOR
                                     (tmp==0x18) ||	// indirect JSM
                                     (tmp==0x1D)) {	// indirect JMP
                                     if ((target >= 0x20) && (target < 0x200)) {
                                     	 if (!isource) {
		                                     if (map_basepage) {
			    	    	     	             if (octal) printf_string(out," [=%06oB]", rom_lower_base_page[target]);
			    	    	     	             else printf_string(out," [=$%04x]", rom_lower_base_page[target]);
		                                     }
		                                     else {
			    	    	     	             if (octal) printf_string(out," [=%06oB]", internals[label-1].value);
			    	    	     	             else printf_string(out," [=$%04x]", internals[label-1].value);
		    	    	     	             }
	    	    	     	             }
	    	    	     	         }
                                }
							}
	    	    	     }
	    	             break;
	    	     case 3: /* register */
	    	     case 4: /* indirect register */
	    	             printf_string(out,"%s", predefined[operand].name);
	    	             if (current->type == 4)
	    	             	printf_string(out,",I");
	    	             break;
	    	     case 5: /* value */
	    	     case 6: /* shift */
	    	     case 7: /* complemented shift */
	    	             if (current->type == 5) {
	    	             	operand++;
	    	             	if (timing == 16) timing = timing + operand * 6;  /* CLR */
	    	             	if (timing == 21) timing = timing + operand * 12; /* XFR */
	    	             	}
	    	             if (current->type == 6) {
	    	             	operand++;
	    	             	timing = timing + operand;
	    	             }
	    	             if (current->type == 7) {                            /* RAL, RBL */
	    	             	operand = 15 - operand;
	    	             	timing = timing - operand;
	    	             }
	    	             printf_string(out,"%d", operand);
	    	             /*
	    	     	     if (octal) {
	    	     	     	printf_string(out,"%oB", operand);
	    	     	     }
	    	     	     else {
	    	     	     	printf_string(out,"$%x", operand);
	    	     	     }
	    	     	     */
	    	     	     break;
	    	     case 8: /* skip */
	    	     case 9: /* skip with set */
	    	     case 10: /* skip with clear */
	    	     case 11: /* complemented skip */
	    	     case 15: /* skip with hold */
	    	             if (current->type == 11) {                           /* RET */
	    	             	 /* operand = ((~operand) & 0x3f) - 32; */
	    	             	 if (operand > 31 ) operand -=64;
				 			 printf_string(out,"%i", operand);
	    	             }
	    	             else { /* all others */
	    	             	 if (operand > 31 ) operand -=64;
	    	    	         target = address + operand;
	    	    	         if (target>65535) target-=65536;
	    	    	         target &= 0xFFFF;
	    	    	         if (target == address)
	    	    	         	printf_string(out,"*");
	    	    	         else {
	    	    	             if (octal)
	    	    	             	printf_string(out,"*%+i", operand);
	    	    	             else
	    	    	             	printf_string(out,"*%+i", operand);

	    	    	             if (current->type == 9)
	    	    	             	printf_string(out,",S");

	    	    	             if (current->type == 10)
	    	    	             	printf_string(out,",C");

	    	    	             label = search_label(internals, internals_size, target);
                           		 if (!isource) {
					     			 if (label)
		                                 printf_string(out," [Lbl_%04i]", label);
		                             else {
	    	    	     	            if (octal)
	    	    	     	            	printf_string(out,"[%06oB]", target);
	    	    	     	            else
	    	    	     	            	printf_string(out,"[$%04x]", target);
		    	    	             }
	    	    	             }
	    	    	         }
	    	    	     }
	    	             break;
	    	     case 12: /* stack register */
	    	     case 13: /* stack register increment */
	    	     case 14: /* stack register decrement */
	    	             printf_string(out,"%s", predefined[operand].name);
	    	             if (current->type == 13)
	    	             	printf_string(out,",I");
	    	             if (current->type == 14)
	    	             	printf_string(out,",D");
	    	             break;
	    	     default: /* unknown */
	    	     	     printf_string(out," ???");
	    	             if (octal)
	    	             	printf_string(out,"[=%06oB]", operand);
	    	     	     else
	    	     	     	printf_string(out,"[=$%04x]", operand);
	    	}
			if (clocks || comments) {
	    		for (j=strlen(out); j<(43+ascii*5+octal*6); j++) out[j]=' ';
	    		out[43+ascii*5+octal*6]=0;
			}
        	if (clocks) {
        		printf_string(out," [%2i]", timing);
        	}
	    	if (comments) {
	    		printf_string(out," ! %s", current->comment);
	    	}
	    	printf("%s\n", out);
	    	break;
	   }
	}
	if (i == op_table_size) {	// unknown instruction, probably data
     		 if (isource) {
     		 	if (octal)
     		 		printf("%sDAT %oB    ! unknown command\n", out, cmd);
     		 	else
     		 		printf("%sDAT %d     ! unknown command\n", out, cmd < 32768 ? cmd : cmd - 65536);
     		 }
     		 else {
	             if (octal) {
	             	printf("%s??? %06oB  ! unknown command\n", out, cmd);
	             }
	     	     else {
	     	     	printf("%s??? $%04x  ! unknown command\n", out, cmd);
	     	     }
     	     }
	}
		
	return;
}

void check_label (unsigned int address, unsigned int cmd)
{
	unsigned int operand;
	Op_code *current = op_table;
	int i, type;
	unsigned int target = 0;

  	for (i=0; i < op_table_size; i++, current++){
	    if ((cmd & current->op_mask) == current->pattern) {
	    	operand = cmd & ~(current->op_mask);
	    	type = get_cmd_type(cmd);
	    	
	    	switch (current->type) {
	    	    case 0:  /* no operand */
	    	             break;
	    	    
	    	    case 1:  /* location */
	    	    case 2:  /* indirect location */
	    	    	     if (operand & 0x400) { /* current page addressing */
	    	    	        operand = operand & 0x3ff;
	    	    	     	if (operand < 512) target = address + operand;
	    	    	     	else target = address - 1024 + operand;
	    	    	     	if (target>65535) target-=65536;
	    	    	     	else if (target<0) target+=65536;
	    	    	        target &= 0xFFFF;
					        if ((target >= start) && (target <= end)) update_label(internals, &internals_size, type, address, target);
					        else {
					        	update_label(externals, &externals_size, type, address, target);
					        }
	    	    	     } else { /* base page addressing */
	    	    	     	operand = operand & 0x3ff;
	    	    	     	if (operand < 512) target = operand;
	    	    	     	else target = 65536 - 1024 + operand;
	    	    	        target &= 0xFFFF;

	    	    	        if (target>31) {
					            if ((target >= start) && (target <= end) && base_page_mode) update_label(internals, &internals_size, type, address, target);
					            else {
					        	update_label(externals, &externals_size, type, address, target);
					            }
					        }
                        	if (map_basepage && (target >= 0x20) && (target < 0x0200)) {
                        		if (*rom_labels[target]) rom_labels_access[target]++;
                        	}
                        	else if (map_basepage && (target >= 0xfe00)) {
                        	 	if (*ram_labels[target-0xfe00]) ram_labels_access[target-0xfe00]++;
					        }
	    	    	     }
			     
	    	             break;
	    	     case 3: /* register */
	    	     case 4: /* indirect register */
	    	             break;
	    	     case 5: /* value */
	    	     case 6: /* shift */
	    	     case 7: /* complemented shift */
	    	     	     break;
	    	     case 8: /* skip */
	    	     case 9: /* skip with set */
	    	     case 10: /* skip with clear */
	    	     case 11: /* complemented skip */
	    	     case 15: /* skip with hold */
	    	             if (current->type != 11) {  /* not RET instruction */
	    	             	if (operand > 31 ) target = address + operand - 64;
	    	             	else target = address + operand;
	    	             	if (target>65535) target-=65536;
	    	    	        target &= 0xFFFF;
			        if ((target >= start) && (target <= end)) update_label(internals, &internals_size, type, address, target);
			        else {
			        	update_label(externals, &externals_size, type, address, target);
			        }
	    	    	     }
	    	             break;
	    	     case 12: /* stack register */
	    	     case 13: /* stack register increment */
	    	     case 14: /* stack register decrement */
	    	             break;
	    	}
	        break;
	   }
	}
		
	return;
}

void init_op_table()
{
	int i = 0;
	op_table_size = 0;
	unsigned int pattern, mask = 0;
	Op_code *index_cmd;
	Str_op_code *index_str;
	
	index_str = str_op_table;
	for (; index_str->opcode; op_table_size++, index_str++);

	if (verbose)
	    printf("Optable size: %i\n", op_table_size);
	
	op_table = malloc(sizeof(struct _op_code) * op_table_size);
	
	if (!op_table) {
		printf("malloc failed.\n");
		exit(-1);
	}
	
	i=0;
	index_str = str_op_table;
	index_cmd = op_table;
	for (; index_str->opcode; index_str++, index_cmd++) {
		index_cmd->opcode = index_str->opcode;
		index_cmd->type = index_str->type;
		index_cmd->timing = index_str->timing;
		index_cmd->comment = index_str->comment;
		pattern = 0;
		for (i=15; i>=0; i--)
			if (index_str->pattern[i] == '1') pattern |= (0x01 << (15-i));
		mask = 0;
		for (i=15; i>=0; i--)
			if (index_str->op_mask[i] == '1') mask |= (0x01 << (15-i));
		index_cmd->pattern = pattern;
		index_cmd->op_mask = mask;
		// printf("%s %u %u %i\n", index_cmd->opcode, index_cmd->pattern, index_cmd->op_mask, index_cmd->type);
	}
	return;
}
