/******************************************************************************

 asm45.h
  
 File revision:	1.0 alpha

 License information:

	Copyright (c) 2007 Ansgar Kueckes

    This program is not free software, however non-commercial use is
    provided without charge and you can redistribute it under the terms
    of the CREATIVE COMMONS PUBLIC LICENSE as published by Creative
    Commons International, either version 3.0 of the Attribution
    NonCommercial NoDerivs License, or any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    CREATIVE COMMONS PUBLIC LICENSE for more details.
    
    You should have received a copy of the CEATIVE COMMONS PUBLIC LICENSE
    along with this program. If not, see
    
    <http://creativecommons.org/licenses/by-nc-nd/3.0/legalcode>.

******************************************************************************/

#define MAX_LABEL 10000
#define MAX_SYMBOLS 10000
#define MAX_LITERALSPACES 100
#define MAX_LEN_SYMBOL 15
#define MAX_LEN_LINE 160


typedef struct _str_op_code {
	char *opcode;
	char *pattern;
	char *op_mask;
	int  type;
	int  timing;
	char *comment;
} Str_op_code;

extern Str_op_code str_op_table[];

typedef struct _op_code {
	char *opcode;
	unsigned int pattern;
	unsigned int op_mask;
	int type;
	int timing;
	char *comment;
} Op_code;

extern Op_code *op_table;
extern int op_table_size;

typedef struct _p_symbol {
	char *name;
	int value;
} P_symbol;

extern int num_predefined;		// number of predefined symbols, will be set in main() section
extern P_symbol predefined[];

typedef enum { 	ABS,			// absolute (base page or indirect)
				REL,			// relative/relocatable (within -512,+511 around program counter)
				EXT				// external (other module or macro)
} Element_type ;

typedef enum { 	ADDRESS,		// machine address (within [0,65535]), size one word
				INTEGER,		// integer (within [-32768,32767]), size one word
				STRING,			// ascii string (even count of bytes), stringsize / 2
				ASCII,			// ascii character (LSB of word), size one word
				REAL,			// full precision bcd floating point number, size four words
				SHORT			// half precision bcd floating point number, size two words
} Value_type ;

typedef struct _element {
	Element_type type;
	Value_type vtype;
	int value;					// used for storing integers & address values
	double bcd;					// used for storing floating point numbers
	char *string;				// used for storing strings
} Element;

typedef struct _symbol {
	Element_type type;
	Value_type vtype;
	char *name;					// symbol identifier
	int value;					// used for storing integers & address values
	double bcd;					// used for storing floating point numbers
	char *string;				// used for storing strings
} Symbol;

typedef struct _literal {
	unsigned int address;		// absolute location of literal
	unsigned char *value;		// binary representation of value (allways word alined)
	int size;					// size of binary representation
	void *next;					// anchor for linked list
} Literal;

typedef struct _literalspace {
	unsigned int address;		// position of literal pool
	int size;					// size of literal pool (in words)
	int used;					// already occupied space in pool
	Literal *list;				// linked list of elements
} Literalspace;


typedef struct _stringlist {
	char *string;
	void *next;
} Stringlist;

extern Symbol symbol_table[];
extern Literalspace literalspace_table[];

extern int symbols_size;
extern int literalspaces_size;

extern int longest_symbol_name;

typedef struct _node {
	unsigned int address;
	void *next;
} Node;

typedef struct _label {
	unsigned int address;
	unsigned int value;
	int refcount;
	int reftype[10];
	Node *ref[10];
} Label;

extern Label internals[MAX_LABEL];
extern Label externals[MAX_LABEL];

extern int internals_size;
extern int externals_size;

extern int num_error_messages;

extern char *error_messages[];	// associated ret value

extern unsigned int *rom_lower_base_page;
extern unsigned int lpu_rom_lower_base_page[];
extern unsigned int ppu_rom_lower_base_page[];
extern char **rom_labels;
extern char *lpu_rom_labels[];
extern char *ppu_rom_labels[];
extern char **ram_labels;
extern char *lpu_ram_labels[];
extern char *ppu_ram_labels[];
extern char **rom_comments;
extern char *lpu_rom_comments[];
extern char *ppu_rom_comments[];
extern char **ram_comments;
extern char *lpu_ram_comments[];
extern char *ppu_ram_comments[];
extern int rom_labels_access[];
extern int ram_labels_access[];


/* switches */
extern int verbose;
extern int octal;
extern int ascii;
extern int comments;
extern int clocks;
extern int offset;
extern int start;
extern int end;
extern int xref;
extern int dump;
extern int disassembly;
extern int strings;
extern int listing;
extern int symbols;
extern int numbers;
extern int base_page_mode;
extern int map_basepage;
extern int isource;


extern unsigned int filesize;

/* common */

void print_error(int line, int err_no);
void printf_string(char *str, char *format, ...);
int get_cmd_type(unsigned int cmd);

/* assembler */

extern int listing_line;

void init_op_table();
void check_label (unsigned int address, unsigned int cmd);
int search_label (Label *list, int size, unsigned int address);
void sort_label_list(Label *list, int size);
void disassemble(int line, unsigned int address, unsigned int cmd, FILE *ofp);
void print_xref(Label *list, int size, int type, int isexternal);

/* disassembler */

void add_symbol(Symbol *symbol);
int assemble_module(FILE *ifp, FILE *override_fp, int *seek_start, int *start, int *input_line);
